@@ -33,6 +33,32 @@
 #include <libinput.h>
 #include <math.h>
 
+#include "libinput-util.h"
+
+struct test_device {
+	const char *name;
+	struct litest_test_device *device;
+} __attribute__((aligned(16)));
+
+#define TEST_DEVICE(name, ...) \
+	static struct litest_test_device _device; \
+	\
+	static void _setup(void) { \
+		struct litest_device *d = litest_create_device(_device.type); \
+		litest_set_current_device(d); \
+	} \
+	\
+	static const struct test_device _test_device \
+		__attribute__ ((used)) \
+		__attribute__ ((section ("test_section"))) = { \
+		name, &_device \
+	}; \
+	static struct litest_test_device _device = { \
+		.setup = _setup, \
+		.shortname = name, \
+		__VA_ARGS__ \
+	};
+
 extern void litest_setup_tests_udev(void);
 extern void litest_setup_tests_path(void);
 extern void litest_setup_tests_pointer(void);
@@ -233,7 +259,12 @@ enum litest_device_type {
 	LITEST_LID_SWITCH_SURFACE3,
 	LITEST_APPLETOUCH,
 	LITEST_GPIO_KEYS,
+	LITEST_IGNORED_MOUSE,
 	LITEST_WACOM_MOBILESTUDIO_PRO_16_PAD,
+	LITEST_THINKPAD_EXTRABUTTONS,
+	LITEST_UCLOGIC_TABLET,
+	LITEST_KEYBOARD_BLADE_STEALTH,
+	LITEST_KEYBOARD_BLADE_STEALTH_VIDEOSWITCH,
 };
 
 enum litest_device_feature {
@@ -266,6 +297,7 @@ enum litest_device_feature {
 	LITEST_TRACKBALL = 1 << 24,
 	LITEST_LEDS = 1 << 25,
 	LITEST_SWITCH = 1 << 26,
+	LITEST_IGNORED = 1 << 27,
 };
 
 /* this is a semi-mt device, so we keep track of the touches that the tests
@@ -305,12 +337,12 @@ struct axis_replacement {
 	double value;
 };
 
+/**
+ * Same as litest_axis_set_value but allows for ranges outside 0..100%
+ */
 static inline void
-litest_axis_set_value(struct axis_replacement *axes, int code, double value)
+litest_axis_set_value_unchecked(struct axis_replacement *axes, int code, double value)
 {
-	litest_assert_double_ge(value, 0.0);
-	litest_assert_double_le(value, 100.0);
-
 	while (axes->evcode != -1) {
 		if (axes->evcode == code) {
 			axes->value = value;
@@ -322,6 +354,18 @@ litest_axis_set_value(struct axis_replacement *axes, int code, double value)
 	litest_abort_msg("Missing axis code %d\n", code);
 }
 
+/**
+ * Takes a value in percent and sets the given axis to that code.
+ */
+static inline void
+litest_axis_set_value(struct axis_replacement *axes, int code, double value)
+{
+	litest_assert_double_ge(value, 0.0);
+	litest_assert_double_le(value, 100.0);
+
+	litest_axis_set_value_unchecked(axes, code, value);
+}
+
 /* A loop range, resolves to:
    for (i = lower; i < upper; i++)
  */
@@ -393,6 +437,13 @@ litest_create_uinput_device_from_description(const char *name,
 					     const struct input_absinfo *abs,
 					     const int *events);
 struct litest_device *
+litest_create(enum litest_device_type which,
+	      const char *name_override,
+	      struct input_id *id_override,
+	      const struct input_absinfo *abs_override,
+	      const int *events_override);
+
+struct litest_device *
 litest_create_device_with_overrides(enum litest_device_type which,
 				    const char *name_override,
 				    struct input_id *id_override,
@@ -556,8 +607,9 @@ litest_keyboard_key(struct litest_device *d,
 		    unsigned int key,
 		    bool is_press);
 
-void litest_lid_action(struct litest_device *d,
-		       enum libinput_switch_state state);
+void litest_switch_action(struct litest_device *d,
+			  enum libinput_switch sw,
+			  enum libinput_switch_state state);
 
 void
 litest_wait_for_event(struct libinput *li);
@@ -627,6 +679,10 @@ litest_is_switch_event(struct libinput_event *event,
 		       enum libinput_switch sw,
 		       enum libinput_switch_state state);
 
+void
+litest_assert_key_event(struct libinput *li, unsigned int key,
+			enum libinput_key_state state);
+
 void
 litest_assert_button_event(struct libinput *li,
 			   unsigned int button,
@@ -671,6 +727,9 @@ litest_timeout_tap(void);
 void
 litest_timeout_tapndrag(void);
 
+void
+litest_timeout_debounce(void);
+
 void
 litest_timeout_softbuttons(void);
 
@@ -701,6 +760,9 @@ litest_timeout_gesture_scroll(void);
 void
 litest_timeout_trackpoint(void);
 
+void
+litest_timeout_tablet_proxout(void);
+
 void
 litest_push_event_frame(struct litest_device *dev);
 
@@ -831,6 +893,15 @@ litest_enable_edge_scroll(struct litest_device *dev)
 	litest_assert_int_eq(status, expected);
 }
 
+static inline bool
+litest_has_clickfinger(struct litest_device *dev)
+{
+	struct libinput_device *device = dev->libinput_device;
+	uint32_t methods = libinput_device_config_click_get_methods(device);
+
+	return methods & LIBINPUT_CONFIG_CLICK_METHOD_CLICKFINGER;
+}
+
 static inline void
 litest_enable_clickfinger(struct litest_device *dev)
 {
@@ -905,6 +976,22 @@ litest_disable_middleemu(struct litest_device *dev)
 	litest_assert_int_eq(status, expected);
 }
 
+static inline bool
+litest_touchpad_is_external(struct litest_device *dev)
+{
+	struct udev_device *udev_device;
+	const char *prop;
+	bool is_external;
+
+	udev_device = libinput_device_get_udev_device(dev->libinput_device);
+	prop = udev_device_get_property_value(udev_device,
+					      "ID_INPUT_TOUCHPAD_INTEGRATION");
+	is_external = prop && streq(prop, "external");
+	udev_device_unref(udev_device);
+
+	return is_external;
+}
+
 #undef ck_assert_double_eq
 #undef ck_assert_double_ne
 #undef ck_assert_double_lt
