@@ -80,6 +80,9 @@ START_TEST(log_handler_invoked)
 {
 	struct libinput *li;
 
+	log_handler_context = NULL;
+	log_handler_called = 0;
+
 	li = libinput_path_create_context(&simple_interface, NULL);
 
 	libinput_log_set_priority(li, LIBINPUT_LOG_PRIORITY_DEBUG);
@@ -89,11 +92,11 @@ START_TEST(log_handler_invoked)
 	libinput_path_add_device(li, "/tmp");
 
 	ck_assert_int_gt(log_handler_called, 0);
-	log_handler_called = 0;
 
 	libinput_unref(li);
 
 	log_handler_context = NULL;
+	log_handler_called = 0;
 }
 END_TEST
 
@@ -101,6 +104,8 @@ START_TEST(log_handler_NULL)
 {
 	struct libinput *li;
 
+	log_handler_called = 0;
+
 	li = libinput_path_create_context(&simple_interface, NULL);
 	libinput_log_set_priority(li, LIBINPUT_LOG_PRIORITY_DEBUG);
 	libinput_log_set_handler(li, NULL);
@@ -108,9 +113,10 @@ START_TEST(log_handler_NULL)
 	libinput_path_add_device(li, "/tmp");
 
 	ck_assert_int_eq(log_handler_called, 0);
-	log_handler_called = 0;
 
 	libinput_unref(li);
+
+	log_handler_called = 0;
 }
 END_TEST
 
@@ -118,6 +124,9 @@ START_TEST(log_priority)
 {
 	struct libinput *li;
 
+	log_handler_context = NULL;
+	log_handler_called = 0;
+
 	li = libinput_path_create_context(&simple_interface, NULL);
 	libinput_log_set_priority(li, LIBINPUT_LOG_PRIORITY_ERROR);
 	libinput_log_set_handler(li, simple_log_handler);
@@ -128,15 +137,15 @@ START_TEST(log_priority)
 	ck_assert_int_eq(log_handler_called, 1);
 
 	libinput_log_set_priority(li, LIBINPUT_LOG_PRIORITY_INFO);
-	/* event0 is usually Lid Switch which prints an info that
-	   we don't handle it */
+	/* event0 exists on any box we care to run the test suite on and we
+	 * currently prints *something* for each device */
 	libinput_path_add_device(li, "/dev/input/event0");
 	ck_assert_int_gt(log_handler_called, 1);
 
-	log_handler_called = 0;
-
 	libinput_unref(li);
+
 	log_handler_context = NULL;
+	log_handler_called = 0;
 }
 END_TEST
 
@@ -148,9 +157,13 @@ axisrange_warning_log_handler(struct libinput *libinput,
 			      const char *format,
 			      va_list args)
 {
+	const char *substr;
+
 	axisrange_log_handler_called++;
 	litest_assert_notnull(format);
-	litest_assert_notnull(strstr(format, "is outside expected range"));
+
+	substr = strstr(format, "is outside expected range");
+	litest_assert_notnull(substr);
 }
 
 START_TEST(log_axisrange_warning)
